// Worker
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include <winsock2.h> // this must be here to avoid windows include order problems

#include "Worker.h"
#include "WorkerWindow.h"
#include "WorkerSettings.h"

// FBuild
#include "Tools/FBuild/FBuildCore/Protocol/Protocol.h"
#include "Tools/FBuild/FBuildCore/Protocol/Server.h"
#include "Tools/FBuild/FBuildCore/WorkerPool/JobQueueRemote.h"
#include "Tools/FBuild/FBuildCore/WorkerPool/WorkerThreadRemote.h"

#include "Core/Env/Env.h"
#include "Core/Network/NetworkStartupHelper.h"
#include "Core/Profile/Profile.h"
#include "Core/Strings/AStackString.h"

// CONSTRUCTOR
//------------------------------------------------------------------------------
Worker::Worker( void * hInstance )
	: m_MainWindow( nullptr )
	, m_ConnectionPool( nullptr )
	, m_NetworkStartupHelper( nullptr )
{
	m_WorkerSettings = FNEW( WorkerSettings );
	m_NetworkStartupHelper = FNEW( NetworkStartupHelper );
	m_JobQueueRemote = FNEW( JobQueueRemote( Env::GetNumProcessors() ) );
	m_MainWindow = FNEW( WorkerWindow( hInstance ) );
	m_ConnectionPool = FNEW( Server );
}

// DESTRUCTOR
//------------------------------------------------------------------------------
Worker::~Worker()
{
	FDELETE m_NetworkStartupHelper;
	FDELETE m_ConnectionPool;
	FDELETE m_MainWindow;
	FDELETE m_JobQueueRemote;
	FDELETE m_WorkerSettings;
}

// Work
//------------------------------------------------------------------------------
int Worker::Work()
{
	// start listening
	if ( m_ConnectionPool->Listen( Protocol::PROTOCOL_PORT ) == false )
	{
		AStackString<> msg;
		msg.Format( "Failed to listen on port %u.  Check port is not in use.", Protocol::PROTOCOL_PORT );
		MessageBox( nullptr, msg.Get(), "FBuild Worker", MB_OK );
		return -1;
	}

	while ( WorkerWindow::Get().WantToQuit() == false )
	{
		UpdateAvailability();

		UpdateUI();

		PROFILE_SYNCHRONIZE

		Sleep( 100 );
	}

	// allow to UI to shutdown
	// the application MUST NOT try to update the UI from this point on
	m_MainWindow->SetAllowQuit();

	m_WorkerBrokerage.SetAvailability( false );

	return 0;
}

// UpdateAvailability
//------------------------------------------------------------------------------
void Worker::UpdateAvailability()
{
	m_IdleDetection.Update();

	WorkerSettings & ws = WorkerSettings::Get();
	uint32_t numCPUsToUse = ws.GetNumCPUsToUse();
	switch( ws.GetMode() )
	{
		case WorkerSettings::WHEN_IDLE:
		{
			if ( m_IdleDetection.IsIdle() == false )
			{
				numCPUsToUse = 0;
			}
			break;
		}
		case WorkerSettings::DEDICATED:
		{
			break; // use all allocated cpus
		}
		case WorkerSettings::DISABLED:
		{
			numCPUsToUse = 0;
			break;
		}
	}

	WorkerThreadRemote::SetNumCPUsToUse( numCPUsToUse );

	m_WorkerBrokerage.SetAvailability( numCPUsToUse > 0 );
}

// UpdateUI
//------------------------------------------------------------------------------
void Worker::UpdateUI()
{
	// title bar
	size_t numConnections = m_ConnectionPool->GetNumConnections();
	AStackString<> status;
	status.Format( "%u Connections", (uint32_t)numConnections );
	m_MainWindow->SetStatus( status.Get() );

	// thread output
	JobQueueRemote & jqr = JobQueueRemote::Get();
	const size_t numWorkers = jqr.GetNumWorkers();
	for ( size_t i=0; i<numWorkers; ++i )
	{
		// get status of worker
		AStackString<> workerStatus;
		AStackString<> hostName;
		bool isIdle;
		jqr.GetWorkerStatus( i, hostName, workerStatus, isIdle );

		// are we syncing tools?
		if ( isIdle )
		{
			AStackString<> statusStr;
			if ( m_ConnectionPool->IsSynchingTool( statusStr ) )
			{
				// show status of synchronization
				workerStatus = statusStr;
			}
		}

		// reflect in UI
		m_MainWindow->SetWorkerState( i, hostName, workerStatus );
	}
}

//------------------------------------------------------------------------------
